# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Rubrics for scoring the output of inference services."""

from . import classification, text
from .base import Rubric

__all__ = ["Rubric", "classification", "text"]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0


def pytest_addoption(parser):
    parser.addoption("--storage_root", action="store", default=None)
    parser.addoption("--test_remote", action="store_true", default=False)
    parser.addoption("--api_endpoint", action="store", default=None)
    parser.addoption("--api_token", action="store", default=None)
    parser.addoption("--api_insecure", action="store_true", default=False)
    parser.addoption("--skip_workflows", action="store_true", default=False)
    parser.addoption("--skip_analyses", action="store_true", default=False)
    parser.addoption("--skip_documentation", action="store_true", default=False)
    parser.addoption("--skip_huggingface", action="store_true", default=False)
    parser.addoption("--skip_errors", action="store_true", default=False)
    parser.addoption("--enable_vllm", action="store_true", default=False)
    parser.addoption("--enable_comparisons", action="store_true", default=False)



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import sys
from typing import Iterable

import pyarrow
import pydantic

from dyff.audit.scoring import Rubric
from dyff.schema.base import int32
from dyff.schema.dataset import ReplicatedItem
from dyff.schema.dataset.arrow import arrow_schema, schema_function


class BlurstCountScoredItem(ReplicatedItem):
    blurstCount: int32() = pydantic.Field(  # type: ignore
        description="Number of times the word 'blurst' is used in the response"
    )


class BlurstCount(Rubric):
    @property
    def name(self) -> str:
        return "dyff.fake.rubric.BlurstCount"

    @property
    @schema_function(arrow_schema(BlurstCountScoredItem))
    def schema(self) -> pyarrow.Schema:
        """The PyArrow schema of the output of applying the Rubric."""

    def apply(
        self, task_data: pyarrow.dataset.Dataset, predictions: pyarrow.dataset.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        print("stdout message")
        print("stderr message", file=sys.stderr)

        for b in predictions.to_batches(
            columns=["_index_", "_replication_", "responses"]
        ):
            batch = []
            for item in b.to_pylist():
                index = item["_index_"]
                replication = item["_replication_"]
                text = item["responses"][0]["text"]
                words = [w.lower() for w in text.split()]
                count = words.count("blurst")
                batch.append(
                    BlurstCountScoredItem(
                        _index_=index, _replication_=replication, blurstCount=count
                    ).dict()
                )
            yield pyarrow.RecordBatch.from_pylist(batch, schema=self.schema)


class RaiseError(Rubric):
    @property
    def name(self) -> str:
        return "dyff.fake.rubric.RaiseError"

    @property
    @schema_function(arrow_schema(BlurstCountScoredItem))
    def schema(self) -> pyarrow.Schema:
        """The PyArrow schema of the output of applying the Rubric."""

    def apply(
        self, task_data: pyarrow.dataset.Dataset, predictions: pyarrow.dataset.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        print("stdout message")
        print("stderr message", file=sys.stderr)
        raise RuntimeError("deliberate error")



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Metrics related to text processing tasks."""

# mypy: disable-error-code="import-untyped"
import pandas
import pyarrow
import pydantic

from dyff.schema.base import DyffSchemaBaseModel, int32
from dyff.schema.dataset.arrow import arrow_schema, schema_function

from ..scoring.text import MatchCriterion, MatchResult
from .base import Metric


def f1_score(df: pandas.DataFrame) -> pandas.Series:
    """F1 score is the harmonic mean of precision and recall.

    :param df: Must contain columns ``precision`` and ``recall``.
    :type df: pandas.DataFrame
    """
    return (2 * df["precision"] * df["recall"]) / (df["precision"] + df["recall"])


class ExtendedPrecisionRecallScore(DyffSchemaBaseModel):
    MatchCriterion: str = pydantic.Field(
        description=":class:`MatchCriterion <dyff.audit.scoring.text.MatchCriterion>` applicable to this record.",
    )
    MatchResultCorrect: int32() = pydantic.Field(
        alias="MatchResult.correct", description="Count of ``correct`` matches."
    )
    MatchResultIncorrect: int32() = pydantic.Field(
        alias="MatchResult.incorrect",
        description="Count of ``incorrect`` matches.",
    )
    MatchResultPartial: int32() = pydantic.Field(
        alias="MatchResult.partial", description="Count of ``partial`` matches."
    )
    MatchResultMissing: int32() = pydantic.Field(
        alias="MatchResult.missing", description="Count of ``missing`` matches."
    )
    MatchResultSpurious: int32() = pydantic.Field(
        alias="MatchResult.spurious", description="Count of ``spurious`` matches."
    )
    possible: int32() = pydantic.Field(description='Number of "possible" matches.')
    actual: int32() = pydantic.Field(description='Number of "actual" matches.')
    precision: float = pydantic.Field(description="Precision score.")
    recall: float = pydantic.Field(description="Recall score.")
    f1_score: float = pydantic.Field(description="F1 score.")


class ExtendedPrecisionRecall(Metric):
    """Compute precision and recall for the extended set of text span matching criteria
    defined in ``dyff.audit.scoring.text``."""

    @property
    def name(self) -> str:
        return "text.ExtendedPrecisionRecall"

    @property
    @schema_function(arrow_schema(ExtendedPrecisionRecallScore))
    def schema(self) -> pyarrow.Schema:
        """The PyArrow schema of the output of applying the Metric."""

    def __call__(self, text_span_matches: pandas.DataFrame) -> pandas.DataFrame:
        def possible(df):
            return sum(
                [
                    df[str(MatchResult.correct)],
                    df[str(MatchResult.incorrect)],
                    df[str(MatchResult.partial)],
                    df[str(MatchResult.missing)],
                ]
            )

        def actual(df):
            return sum(
                [
                    df[str(MatchResult.correct)],
                    df[str(MatchResult.incorrect)],
                    df[str(MatchResult.partial)],
                    df[str(MatchResult.spurious)],
                ]
            )

        def precision(row):
            if row["MatchCriterion"] in [MatchCriterion.strict, MatchCriterion.exact]:
                return row[str(MatchResult.correct)] / row["actual"]
            else:
                return (
                    row[str(MatchResult.correct)] + 0.5 * row[str(MatchResult.partial)]
                ) / row["actual"]

        def recall(row):
            if row["MatchCriterion"] in [MatchCriterion.strict, MatchCriterion.exact]:
                return row[str(MatchResult.correct)] / row["possible"]
            else:
                return (
                    row[str(MatchResult.correct)] + 0.5 * row[str(MatchResult.partial)]
                ) / row["possible"]

        keep_columns = ["MatchCriterion"] + [str(r) for r in MatchResult]
        sums = (
            text_span_matches.filter(keep_columns)
            .groupby("MatchCriterion")
            .sum()
            .reset_index()
        )
        sums["possible"] = possible(sums)
        sums["actual"] = actual(sums)
        sums["precision"] = sums.apply(precision, axis=1)
        sums["recall"] = sums.apply(recall, axis=1)
        sums["f1_score"] = f1_score(sums)
        return sums



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from dyff.schema.dataset import arrow
from dyff.schema.platform import (
    MeasurementLevel,
    MethodImplementationKind,
    MethodOutputKind,
)

from .. import dynamic_import
from ..analysis import AnalysisContext
from ..scoring import Rubric


def run_python_function():
    ctx = AnalysisContext()
    implementation = ctx.analysis.method.implementation
    if (
        implementation.kind != MethodImplementationKind.PythonFunction
        or implementation.pythonFunction is None
    ):
        raise ValueError("expected method.implementation as PythonFunction")
    output = ctx.analysis.method.output
    if output.kind != MethodOutputKind.Measurement or output.measurement is None:
        raise ValueError("expected method.output as Measurement")

    function = dynamic_import.symbol(implementation.pythonFunction.fullyQualifiedName)
    input_data = {i: ctx.open_input_dataset(i) for i in ctx.inputs}
    output_generator = function(ctx.arguments, **input_data)
    output_schema = arrow.decode_schema(output.measurement.schema_.arrowSchema)

    # TODO: We're not doing anything with the DataViews yet

    arrow.write_dataset(
        output_generator,
        output_path=ctx.output_path,
        feature_schema=output_schema,
    )


def run_python_rubric():
    ctx = AnalysisContext()
    implementation = ctx.analysis.method.implementation
    if (
        implementation.kind != MethodImplementationKind.PythonRubric
        or implementation.pythonRubric is None
    ):
        raise ValueError("expected method implemented as PythonRubric")
    output = ctx.analysis.method.output
    if output.kind != MethodOutputKind.Measurement or output.measurement is None:
        raise ValueError("expected method.output as Measurement")
    if output.measurement.level != MeasurementLevel.Instance:
        raise ValueError("expected measurement.level == Instance")
    if ctx.arguments:
        raise ValueError("Rubrics do not accept arguments")

    fqn = implementation.pythonRubric.fullyQualifiedName
    try:
        rubric = dynamic_import.instantiate(fqn)
    except ImportError:
        # Backwards compatibility
        rubric = dynamic_import.instantiate(f"dyff.audit.scoring.{fqn}")
    if not isinstance(rubric, Rubric):
        raise ValueError(f"{fqn} is not a Rubric subclass")
    output_schema = arrow.decode_schema(output.measurement.schema_.arrowSchema)

    task_data = ctx.open_input_dataset("task_data")
    predictions = ctx.open_input_dataset("predictions")

    # TODO: We're not doing anything with the DataViews yet

    arrow.write_dataset(
        rubric.apply(task_data, predictions),
        output_path=str(ctx.output_path),
        feature_schema=output_schema,
    )



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import sys
from typing import Optional

from dyff.schema.dataset import arrow

from .. import dynamic_import
from ..scoring import Rubric


def run_python_rubric(
    *,
    rubric: str,
    dataset_path: str,
    evaluation_path: str,
    output_path: str,
    modules: Optional[list[str]] = None,
):
    # Add extension modules to import path
    if modules:
        for module in modules:
            sys.path.append(module)

    try:
        rubric = dynamic_import.instantiate(rubric)
    except ImportError:
        # Backwards compatibility
        rubric = dynamic_import.instantiate(f"dyff.audit.scoring.{rubric}")
    if not isinstance(rubric, Rubric):
        raise ValueError(f"{rubric} is not a Rubric subclass")

    task_data = arrow.open_dataset(dataset_path)
    outputs_data = arrow.open_dataset(evaluation_path)

    # TODO: We're not doing anything with the DataViews yet

    arrow.write_dataset(
        rubric.apply(task_data, outputs_data),
        output_path=output_path,
        feature_schema=rubric.schema,
    )



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os
import sys
import tempfile
from pathlib import Path

from dyff.client import Client


def main():
    endpoint = os.environ.get("DYFF_API_ENDPOINT")
    token = os.environ["DYFF_API_TOKEN"]
    insecure = os.environ.get("DYFF_API_INSECURE") == "1"
    dyffapi = Client(api_key=token, endpoint=endpoint, insecure=insecure)

    dataset_id = sys.argv[1]

    with tempfile.TemporaryDirectory() as tmp:
        dyffapi.datasets.download(dataset_id, Path(tmp) / "data")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Metrics are summaries of the scores generated from Rubrics."""

from . import text
from .base import Metric

__all__ = ["Metric", "text"]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

from pathlib import Path
from typing import Iterable, Optional

import pyarrow

from dyff.client import Client
from dyff.schema import ids
from dyff.schema.dataset import arrow
from dyff.schema.platform import InferenceSession

from .scoring import Rubric


def local_evaluation(
    client: Client,
    session: InferenceSession,
    *,
    input_dataset_path: Path | str,
    output_dataset_path: Path | str,
    replications: int = 1,
    id: Optional[str] = None,
) -> str:
    """Emulate an Evaluation workflow by feeding data from a local Arrow dataset to an
    InferenceSession running on the Dyff platform.

    The output dataset will have the same schema as the outputs from an
    Evaluation run on the platform, including fields added by the platform
    -- ``_index_``, ``_replication_``, etc.

    The input dataset must be compatible with the canonical Dyff Platform
    dataset schema for the appropriate inference task.

    :param client: A Dyff API client with permission to call the
        ``inferencesessions.token()`` and ``inferencesessions.client()``
        endpoints on the given ``InferenceSession``.
    :type client: dyff.client.Client
    :param session: A record describing a running ``InferenceSession``.
    :type session: dyff.schema.platform.InferenceSession
    :keyword input_dataset_path: The root directory of an Arrow dataset on
        the local filesystem.
    :type input_dataset_path: Path | str
    :keyword output_dataset_path: The directory where the Arrow output dataset
        should be created. A subdirectory named with the ID of the simulated
        evaluation will be created.
    :type output_dataset_path: Path | str
    :keyword replications: The number of replications to run. Equivalent to the
        ``EvaluationCreateRequest.replications`` parameter.
    :type replications: int
    :keyword id: If specified, use this ID for the evaluation. Otherwise,
        generate a new ID.
    :type id: str
    :returns: An ID for the simulated evaluation, either the ID provided as
        an argument, or a newly-generated one. This will not correspond to
        an entity in the Dyff datastore, but it can be used to derive the IDs
        of replications in the output dataset.
    :rtype: str
    """
    session_token = client.inferencesessions.token(session.id)
    interface = session.inferenceService.interface
    inference_client = client.inferencesessions.client(
        session.id, session_token, interface=interface
    )
    evaluation_id = id or ids.generate_entity_id()
    replication_ids = [
        ids.replication_id(evaluation_id, i) for i in range(replications)
    ]
    feature_schema = arrow.decode_schema(interface.outputSchema.arrowSchema)
    partition_schema = arrow.subset_schema(feature_schema, ["_replication_"])

    def output_generator() -> Iterable[pyarrow.RecordBatch]:
        input_dataset = arrow.open_dataset(str(input_dataset_path))
        for input_batch in input_dataset.to_batches():
            output_batch: list[pyarrow.RecordBatch] = []
            for item in input_batch.to_pylist():
                index = item["_index_"]
                for replication in replication_ids:
                    responses = inference_client.infer(item)
                    for i, response in enumerate(responses):
                        response["_response_index_"] = i
                    response_record = {
                        "_replication_": replication,
                        "_index_": index,
                        "responses": responses,
                    }
                    output_batch.append(response_record)
            yield pyarrow.RecordBatch.from_pylist(output_batch, schema=feature_schema)

    evaluation_output_path = Path(output_dataset_path) / evaluation_id
    evaluation_output_path.mkdir()
    arrow.write_dataset(
        output_generator(),
        output_path=str(evaluation_output_path),
        feature_schema=feature_schema,
        partition_schema=partition_schema,
    )

    return evaluation_id


def local_report(
    rubric: Rubric,
    *,
    input_dataset_path: Path | str,
    output_dataset_path: Path | str,
    report_dataset_path: Path | str,
):
    """Emulate a Report workflow on local data.

    You will need the Arrow datasets of inputs and outputs to an Evaluation
    workflow. You can emulate an Evaluation locally with ``local_evaluation()``.

    :param rubric: The Rubric to apply.
    :type rubric: dyff.audit.scoring.Rubric
    :keyword input_dataset_path: The root directory of the Arrow dataset
        containing the inputs to an evaluation.
    :type input_dataset_path: Path | str
    :keyword output_dataset_path: The root directory of the Arrow dataset
        containing the outputs of the evaluation.
    :type output_dataset_path: Path | str
    :keyword report_dataset_path: The directory where the Arrow dataset of
        report outputs should be created. A subdirectory named with the ID of
        the simulated report will be created.
    :type report_dataset_path: Path | str
    :returns: An ID for the simulated report. This will not correspond to
        an entity in the Dyff datastore.
    :rtype: str
    """

    def output_generator() -> Iterable[pyarrow.RecordBatch]:
        input_dataset = arrow.open_dataset(str(input_dataset_path))
        output_dataset = arrow.open_dataset(str(output_dataset_path))
        yield from rubric.apply(input_dataset, output_dataset)

    report_id = ids.generate_entity_id()
    report_output_path = Path(report_dataset_path) / report_id
    report_output_path.mkdir()
    arrow.write_dataset(
        output_generator(),
        output_path=str(report_output_path),
        feature_schema=rubric.schema,
    )

    return report_id


__all__ = [
    "local_evaluation",
    "local_report",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import random

import pyarrow
import pydantic

from dyff.schema.dataset import Item, arrow


class Text(Item):
    text: str = pydantic.Field(description="Text data")


def main():
    num_rows = 10_000
    words_per_row = 10
    batch_size = 100
    arrow_schema = arrow.arrow_schema(Text)

    with open("tests/words.txt", "r") as fin:
        words = fin.readlines()

    def g():
        batch = []
        for i in range(num_rows):
            text = " ".join(random.choices(words, k=words_per_row))
            batch.append(Text(_index_=i, text=text).model_dump(mode="python"))
            if len(batch) == batch_size:
                yield pyarrow.RecordBatch.from_pylist(batch, schema=arrow_schema)
                batch = []
        if batch:
            yield pyarrow.RecordBatch.from_pylist(batch, schema=arrow_schema)

    arrow.write_dataset(
        g(), output_path="tests/data/dataset", feature_schema=arrow_schema
    )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import nbconvert
import nbformat

# mypy: disable-error-code="import-untyped"
from bs4 import BeautifulSoup

from dyff.audit.analysis.context import AnalysisContext
from dyff.schema.platform import MethodImplementationKind, MethodOutputKind


def _bs4_parse(html: str) -> BeautifulSoup:
    return BeautifulSoup(html, features="lxml")


def run_jupyter_notebook() -> None:
    ctx = AnalysisContext()

    implementation = ctx.analysis.method.implementation
    if (
        implementation.kind != MethodImplementationKind.JupyterNotebook
        or implementation.jupyterNotebook is None
    ):
        raise ValueError("expected method.implementation as JupyterNotebook")
    output = ctx.analysis.method.output
    if output.kind != MethodOutputKind.SafetyCase or output.safetyCase is None:
        raise ValueError("expected method.output as SafetyCase")

    notebook_path = (
        ctx.local_storage_root
        / implementation.jupyterNotebook.notebookModule
        / implementation.jupyterNotebook.notebookPath
    )
    with open(str(notebook_path), "r") as fin:
        notebook = nbformat.read(fin, as_version=4)

    resources: dict = {}
    clear_output_preprocessor = nbconvert.preprocessors.ClearOutputPreprocessor()
    notebook, resources = clear_output_preprocessor.preprocess(notebook, resources)

    execute_preprocessor = nbconvert.preprocessors.ExecutePreprocessor()
    notebook, resources = execute_preprocessor.preprocess(notebook, resources)

    html_exporter = nbconvert.HTMLExporter(
        exclude_input=True, exclude_input_prompt=True, exclude_output_prompt=True
    )
    html_body, resources = html_exporter.from_notebook_node(notebook, resources)

    # Note: It is arguably cleaner to do this step in a custom Jupyter exporter,
    # but this way is much easier to distribute and it works fine as long as
    # the number of changes needed is minimal.
    soup = _bs4_parse(html_body)
    # Insert tailwind CSS in <head> by reference
    head = soup.find("head")
    tailwind_css = soup.new_tag(
        "script",
        src="https://cdn.tailwindcss.com",
    )
    body_style = _bs4_parse(
        """
        <style type="text/css">
            @import url('https://fonts.googleapis.com/css2?family=IBM+Plex+Serif:wght@400;700&display=swap');

            body {
                padding: 0 !important;
                margin: 0 !important;
            }

            main {
                padding: 0;
                display: grid;
                grid-template-columns: repeat(10, 1fr);
            }

            .jp-Cell {
                @media (width < 768px) {
                    grid-column: span 10 !important;
                }
                @media (768px <= width < 1000px) {
                    grid-column: 2 / span 8 !important;
                }
                @media (1000px <= width <= 1500px) {
                    grid-column: 3 / span 6 !important;
                }
                @media (1500px < width) {
                    grid-column: 4 / span 4 !important;
                }
            }

            .jp-Cell:has(.jp-all-columns) {
                grid-column: span 10 !important;
                padding: 0 !important;
            }

            .jp-RenderedHTML:not(:has(.jp-all-columns)) {
                padding-left: 1rem !important;
                padding-right: 1rem !important;
            }

            .jp-RenderedHTML {
                padding: 0 !important;
            }

            .jp-mod-noOutputs .jp-mod-noInput .jp-OutputArea-prompt .jp-InputArea-prompt {
                display: none !important;
            }

            :root {
                --jp-notebook-padding: 0;
                --jp-cell-padding: 10px;

                --jp-content-line-height: 1.6;
                --jp-content-font-scale-factor: 1.2
                --jp-content-font-size0: 1rem;
                --jp-content-font-size1: 1.125rem;
                --jp-content-font-size2: 1.25rem;
                --jp-content-font-size3: 1.5rem;
                --jp-content-font-size4: 2rem;
                --jp-content-font-size5: 3rem;


                --jp-ui-font-scale-factor: 1.2;
                --jp-ui-font-size0: 1rem;
                --jp-ui-font-size1: 1.125rem; /* Base font size */
                --jp-ui-font-size2: 1.25rem;
                --jp-ui-font-size3: 1.5rem;

                --jp-code-font-size: 1.125rem;
                --jp-code-line-height: 1.6;
                --jp-code-font-family-default: menlo, consolas, 'DejaVu Sans Mono', monospace;
                --jp-code-font-family: var(--jp-code-font-family-default);

                --jp-content-font-family: system-ui, -apple-system, blinkmacsystemfont,
                    'Segoe UI', helvetica, arial, sans-serif, 'Apple Color Emoji',
                    'Segoe UI Emoji', 'Segoe UI Symbol';
                --jp-ui-font-family: system-ui, -apple-system, blinkmacsystemfont, 'Segoe UI',
                    helvetica, arial, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji',
                    'Segoe UI Symbol';
            }
        </style>
        """
    )

    date_script = _bs4_parse(
        """
        <script type="text/javascript">
            window.onload = (_) => {
                const dateEl = document.querySelector("#date_berry");
                if (!dateEl) return;
                const isoDateStr = dateEl.innerHTML;
                let styledDateStr = isoDateStr;
                try {
                    const isoDateObj = new Date(isoDateStr);
                    styledDateStr = isoDateObj.toLocaleDateString();
                } catch (_) {
                    // Pass
                }
                dateEl.innerHTML = styledDateStr;
            };
        </script>
        """
    )
    head.append(tailwind_css)
    head.append(body_style)
    head.append(date_script)
    html_body = str(soup)

    ctx.output_path.mkdir(parents=True, exist_ok=True)
    with open(ctx.output_path / "index.html", "w") as fout:
        fout.write(html_body)



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import List

from dyff.schema.dataset.text import TaggedSpan


def token_tags_to_spans(
    text: str, tokens: List[str], tags: List[str]
) -> List[TaggedSpan]:
    """Computes the list of TaggedSpans corresponding to the tagged tokens in the
    text."""
    spans: List[TaggedSpan] = []
    i = 0
    start = None
    end = None
    current_tag = None

    def finish_span():
        nonlocal spans, start, end, current_tag

        if current_tag is not None:
            spans.append(TaggedSpan(start=start, end=end, tag=current_tag))
        start = None
        end = None
        current_tag = None

    for token_index, token in enumerate(tokens):
        tag = tags[token_index]
        if tag == "O":
            finish_span()
        elif tag.startswith("B-"):
            finish_span()
            current_tag = tag[2:]
        elif tag.startswith("I-"):
            assert tag[2:] == current_tag

        while text[i] != token[0]:
            i += 1
        if start is None:
            start = i
        for c in token:
            assert text[i] == c
            i += 1
        end = i
    finish_span()

    return spans


def visualize_spans(text: str, spans: List[TaggedSpan], *, width: int = 80):
    """Print lines of text with lines representing NER spans aligned underneath.

    Example output::

      My name is Alice and I live in Alaska.
                 PPPPP               LLLLLL
    """
    start = 0
    next_span = 0
    while start < len(text):
        span_text: List[str] = []
        end = min(start + 80, len(text))
        print(text[start:end])
        s = start
        while next_span < len(spans):
            span = spans[next_span]
            span_start = max(span.start, start)
            span_end = min(span.end, end)
            if span_start < end:
                span_text.extend("." * (span_start - s))
                span_text.extend(span.tag[0] * (span_end - span_start))
            s = span_end
            if span.end >= end:
                break
            next_span += 1
        if len(span_text) < width:
            span_text.extend("." * ((end - start) - len(span_text)))
        print("".join(span_text))
        start += width


__all__ = [
    "token_tags_to_spans",
    "visualize_spans",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path

from dyff.client import Client


def main():
    account = "test"
    endpoint = os.environ.get("DYFF_API_ENDPOINT")
    token = os.environ["DYFF_API_TOKEN"]
    insecure = os.environ.get("DYFF_API_INSECURE") == "1"
    dyffapi = Client(api_key=token, endpoint=endpoint, insecure=insecure)

    dataset_dir = Path("tests") / "data" / "dataset"
    dataset = dyffapi.datasets.create_arrow_dataset(
        dataset_dir, account=account, name="dataset"
    )
    dyffapi.datasets.upload_arrow_dataset(dataset, dataset_dir)
    print(f"dataset: {dataset.id}")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import importlib

import pytest


@pytest.mark.parametrize(
    "module_name",
    [
        "dyff.audit",
        "dyff.audit.analysis",
        "dyff.audit.analysis.context",
        "dyff.audit.analysis.jupyter",
        "dyff.audit.analysis.legacy",
        "dyff.audit.analysis.python",
        "dyff.audit.data",
        "dyff.audit.data.text",
        "dyff.audit.dynamic_import",
        "dyff.audit.local",
        "dyff.audit.local.platform",
        "dyff.audit.metrics.base",
        "dyff.audit.metrics.text",
        "dyff.audit.metrics",
        "dyff.audit.scoring.base",
        "dyff.audit.scoring.example",
        "dyff.audit.scoring.text",
        "dyff.audit.scoring",
        "dyff.audit.scoring.classification",
        "dyff.audit.workflows",
    ],
)
def test_import_module(module_name):
    importlib.import_module(module_name)
